/*:
# Bridge: Real World Example

You can go back to the [conceptual example](@previous) if you want to review the concepts.
 */
import UIKit

private protocol SharingSupportable {
    /// Abstraction
    func accept(service: SharingService)
    func update(content: Content)
}

class BaseViewController: UIViewController, SharingSupportable {
    fileprivate var shareService: SharingService?

    func update(content: Content) {
        /// ...updating UI and showing a content...
        /// ...
        /// ... then, a user will choose a content and trigger an event
        print("\(description): User selected a \(content) to share")
        /// ...
        shareService?.share(content: content)
    }

    func accept(service: SharingService) {
        shareService = service
    }
}

class PhotoViewController: BaseViewController {
    /// Custom UI and features

    override var description: String {
        return "PhotoViewController"
    }
}

class FeedViewController: BaseViewController {
    /// Custom UI and features

    override var description: String {
        return "FeedViewController"
    }
}

protocol SharingService {
    /// Implementation
    func share(content: Content)
}

class FaceBookSharingService: SharingService {
    func share(content: Content) {
        /// Use FaceBook API to share a content
        print("Service: \(content) was posted to the Facebook")
    }
}

class InstagramSharingService: SharingService {
    func share(content: Content) {
        /// Use Instagram API to share a content
        print("Service: \(content) was posted to the Instagram", terminator: "\n\n")
    }
}

protocol Content: CustomStringConvertible {
    var title: String { get }
    var images: [UIImage] { get }
}

struct FoodDomainModel: Content {
    var title: String
    var images: [UIImage]
    var calories: Int

    var description: String {
        return "Food Model"
    }
}

class BridgeRealWorld_TestCase {
    func testBridgeRealWorld() {
        print("Client: Pushing Photo View Controller...")
        push(PhotoViewController())

        print()

        print("Client: Pushing Feed View Controller...")
        push(FeedViewController())
    }

    fileprivate func push(_ container: SharingSupportable) {
        let instagram = InstagramSharingService()
        let facebook = FaceBookSharingService()

        container.accept(service: instagram)
        container.update(content: foodModel)

        container.accept(service: facebook)
        container.update(content: foodModel)
    }

    var foodModel: Content {
        return FoodDomainModel(title: "This food is so various and delicious!",
                               images: [UIImage(), UIImage()],
                               calories: 47)
    }
}

let testCase = BridgeRealWorld_TestCase()
testCase.testBridgeRealWorld()
